import time
from typing import List
from logging import Logger
from confluent_kafka import Consumer

from . import default_logger


class KafkaConsumer:
    def __init__(self, topics: List[str], config: dict, handler, logger: Logger = None):
        self.consumer = Consumer(config)
        self.topics = topics
        self.handler = handler
        if logger is None:
            logger = default_logger
        self.logger = logger

    def consume(self):
        try:
            self.consumer.subscribe(self.topics)
            while True:
                msg = self.consumer.poll(timeout=1.0)
                if msg is None or msg.error():
                    continue

                decoded_msg = msg.value().decode("utf-8")
                self.logger.info(
                    f"New message received: {decoded_msg} from topic {msg.topic()}"
                )
                self.handler(decoded_msg)
        finally:
            self.consumer.close()


class MockConsumer:
    def __init__(self, wait_time, runs, handler):
        self.wait_time = wait_time
        self.runs = runs
        self.handler = handler

    def consume(self):
        for _ in range(self.runs):
            self.handler("")
            time.sleep(self.wait_time)



from .interface import Comm, LogLevel
from .dispatcher import Dispatcher
from .topic import TopicComm
from .log import LogComm
from .monitoring import MonitoringComm
from .prometheus import PrometheusComm



import logging
from abc import ABC, abstractmethod
from enum import Enum


class LogLevel(Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL
    SUCCESS = 1

    def to_logging_level(self) -> int:
        if self == LogLevel.SUCCESS:
            return logging.INFO

        return self.value


class Comm(ABC):
    @abstractmethod
    def send(*args, **kwargs):
        pass



from .interface import Comm
from eo4eu_comm_utils.kafka import KafkaProducer


class TopicComm(Comm):
    def __init__(self, producer: KafkaProducer, topic: str, key: str):
        self.producer = producer
        self.topic = topic
        self.key = key

    def send(self, message: str = "", *args, **kwargs):
        self.producer.send_message(
            key = self.key,
            msg = message,
            topic = self.topic
        )



import traceback
import logging
from .interface import Comm, LogLevel


class LogComm(Comm):
    def __init__(self, logger: logging.Logger, print_traceback: bool = False):
        self.logger = logger
        self.print_traceback = print_traceback

    def send(self, level: LogLevel, msg: str = "", *args, **kwargs):
        logging_level = level.to_logging_level()
        self.logger.log(
            logging_level,
            msg,
            exc_info = self.print_traceback and (logging_level >= logging.ERROR)
        )



from typing import Self

from .interface import Comm, LogLevel


class Dispatcher:
    def __init__(
        self,
        comms: dict[str,Comm],
        groups: dict[str,list[str]]|None = None,
        callback = None,
        selection: list[str]|None = None
    ):
        self._comms = comms
        if groups is None:
            groups = {}
        self._groups = groups
        if callback is None:
            callback = lambda this, *args, **kwargs: None
        self._callback = callback
        self._selection = selection

    def add_comm(self, name: str, comm: Comm) -> Self:
        self._comms[name] = comm
        return self

    def add_group(self, name: str, sub_names: str) -> Self:
        result = []
        for sub_name in sub_names:
            if sub_name in self._comms:
                result.append(sub_name)
            elif sub_name in self._groups:
                result.extend(self._groups[sub_name])

        self._groups[name] = result
        return self

    def _with_selection(self, selection: list[str]|None) -> Self:
        return Dispatcher(
            comms = self._comms,
            groups = self._groups,
            callback = self._callback,
            selection = selection
        )

    def __getattr__(self, name: str) -> Self:
        if name in self._groups:
            return self._with_selection(self._groups[name])

        if name in self._comms:
            return self._with_selection([name])

        return self._with_selection([])

    def all(self) -> Self:
        return self._with_selection(None)

    def len(self) -> int:
        if self._selection is None:
            return len(self._comms)

        return len(self._selection)

    def send(self, *args, **kwargs):
        self._callback(self, *args, **kwargs)
        selection = self._comms.keys() if self._selection is None else self._selection

        for name in selection:
            self._comms[name].send(*args, **kwargs)

    def add(self, *args, **kwargs):
        self.send(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self.send(LogLevel.DEBUG, *args, **kwargs)

    def info(self, *args, **kwargs):
        self.send(LogLevel.INFO, *args, **kwargs)

    def warning(self, *args, **kwargs):
        self.send(LogLevel.WARNING, *args, **kwargs)

    def error(self, *args, **kwargs):
        self.send(LogLevel.ERROR, *args, **kwargs)

    def critical(self, *args, **kwargs):
        self.send(LogLevel.CRITICAL, *args, **kwargs)

    def success(self, *args, **kwargs):
        self.send(LogLevel.SUCCESS, *args, **kwargs)



import prometheus_client
from enum import Enum

from .interface import Comm


class PrometheusComm(Comm):
    def __init__(self, input: dict[Enum,tuple[str,str]]):
        prometheus_client.start_http_server(8000)
        self._metrics = {
            kind: prometheus_client.Counter(name, desc)
            for kind, (name, desc) in input.items()
        }

    def send(self, *kinds: Enum, value: int = 1, **kwargs):
        for kind in kinds:
            self._metrics[kind].inc(value)



import json
from datetime import datetime

from .interface import Comm, LogLevel
from eo4eu_comm_utils.kafka.producer import KafkaProducer


class MonitoringComm(Comm):
    def __init__(
        self,
        producer: KafkaProducer,
        namespace: str,
        source: str,
        prefix: str,
        level_dict: dict[LogLevel,str]|None = None,
        **kwargs
    ):
        self.producer = producer
        self.namespace = namespace
        self.source = source
        self.prefix = prefix
        if level_dict is None:
            level_dict = {
                LogLevel.DEBUG: ("INFO", "Info"),
                LogLevel.INFO: ("INFO", "Info"),
                LogLevel.WARNING: ("WARNING", "Error"),
                LogLevel.ERROR: ("WARNING", "Error"),
                LogLevel.CRITICAL: ("CRITICAL", "Error"),
                LogLevel.SUCCESS: ("SUCCESS", "Info"),
            }
        self.level_dict = level_dict
        self.kwargs = kwargs

    def send(
        self,
        level: LogLevel,
        description: str,
        *args,
        description_business: str = None,
        **kwargs
    ):
        if description_business is None:
            description_business = description

        level_name, msg_name = self.level_dict[level]
        message = {
            "source": self.source,
            "name": f"{self.prefix}{msg_name}",
            "level": level_name,
            "description": description,
            "description_business": description_business,
            "timestamp": datetime.utcnow().isoformat(),
            "extra": {
                "namespace": self.namespace,
                **(self.kwargs | kwargs),
            }
        }

        self.producer.send_message(
            key = self.source,
            msg = json.dumps(message),
            topic = "monitoring.notify"
        )



from logging import Logger
from confluent_kafka import Producer

from . import default_logger


class KafkaProducer:
    def __init__(self, topic: str, config: dict, logger: Logger = None):
        # self.topic = "system.dsl"
        self.topic = topic
        self.producer = Producer(config)
        if logger is None:
            logger = default_logger
        self.logger = logger

    def set_topic(self, topic: str):
        self.topic = topic

    def send_message(self, key: str, msg: str, topic: str = None):
        def _acked(err, msg):
            if err is not None:
                self.logger.error(f"Failed to deliver message: {str(msg)}: {str(err)}")
            else:
                self.logger.info(
                    f"Message produced: {str(msg.value())} on topic {self._topic}"
                )

        if topic is None:
            topic = self.topic

        self.producer.produce(topic, key=key, value=msg, callback=_acked)
        self.producer.flush()


class MockProducer:
    def __init__(self, topic: str = "", logger: Logger = None):
        self.topic = topic
        if logger is None:
            logger = default_logger
        self.logger = logger

    def set_topic(self, topic: str):
        self.topic = topic

    def send_message(self, key: str, msg: str, topic: str = None):
        if topic is None:
            topic = self.topic

        self.logger.info(f"Producer, key: {key}, topic: {topic}")
        self.logger.info(msg)
